@@ -1,28 +1,43 @@
-from __future__ import unicode_literals
-
+import ipaddress
 import math
 import re
 import uuid
-import warnings
 
 try:
     import email_validator
 except ImportError:
     email_validator = None
-try:
-    import ipaddress
-except ImportError:
-    ipaddress = None
-
-from wtforms.compat import string_types, text_type
 
 __all__ = (
-    'DataRequired', 'data_required', 'Email', 'email', 'EqualTo', 'equal_to',
-    'IPAddress', 'ip_address', 'InputRequired', 'input_required', 'Length',
-    'length', 'NumberRange', 'number_range', 'Optional', 'optional',
-    'Required', 'required', 'Regexp', 'regexp', 'URL', 'url', 'AnyOf',
-    'any_of', 'NoneOf', 'none_of', 'MacAddress', 'mac_address', 'UUID',
-    'ValidationError', 'StopValidation'
+    "DataRequired",
+    "data_required",
+    "Email",
+    "email",
+    "EqualTo",
+    "equal_to",
+    "IPAddress",
+    "ip_address",
+    "InputRequired",
+    "input_required",
+    "Length",
+    "length",
+    "NumberRange",
+    "number_range",
+    "Optional",
+    "optional",
+    "Regexp",
+    "regexp",
+    "URL",
+    "url",
+    "AnyOf",
+    "any_of",
+    "NoneOf",
+    "none_of",
+    "MacAddress",
+    "mac_address",
+    "UUID",
+    "ValidationError",
+    "StopValidation",
 )
 
 
@@ -30,7 +45,8 @@ class ValidationError(ValueError):
     """
     Raised when a validator fails to validate its input.
     """
-    def __init__(self, message='', *args, **kwargs):
+
+    def __init__(self, message="", *args, **kwargs):
         ValueError.__init__(self, message, *args, **kwargs)
 
 
@@ -42,11 +58,12 @@ class StopValidation(Exception):
     called. If raised with a message, the message will be added to the errors
     list.
     """
-    def __init__(self, message='', *args, **kwargs):
+
+    def __init__(self, message="", *args, **kwargs):
         Exception.__init__(self, message, *args, **kwargs)
 
 
-class EqualTo(object):
+class EqualTo:
     """
     Compares the values of two fields.
 
@@ -57,6 +74,7 @@ class EqualTo(object):
         interpolated with `%(other_label)s` and `%(other_name)s` to provide a
         more helpful error.
     """
+
     def __init__(self, fieldname, message=None):
         self.fieldname = fieldname
         self.message = message
@@ -64,21 +82,27 @@ class EqualTo(object):
     def __call__(self, form, field):
         try:
             other = form[self.fieldname]
-        except KeyError:
-            raise ValidationError(field.gettext("Invalid field name '%s'.") % self.fieldname)
-        if field.data != other.data:
-            d = {
-                'other_label': hasattr(other, 'label') and other.label.text or self.fieldname,
-                'other_name': self.fieldname
-            }
-            message = self.message
-            if message is None:
-                message = field.gettext('Field must be equal to %(other_name)s.')
+        except KeyError as exc:
+            raise ValidationError(
+                field.gettext("Invalid field name '%s'.") % self.fieldname
+            ) from exc
+        if field.data == other.data:
+            return
 
-            raise ValidationError(message % d)
+        d = {
+            "other_label": hasattr(other, "label")
+            and other.label.text
+            or self.fieldname,
+            "other_name": self.fieldname,
+        }
+        message = self.message
+        if message is None:
+            message = field.gettext("Field must be equal to %(other_name)s.")
+
+        raise ValidationError(message % d)
 
 
-class Length(object):
+class Length:
     """
     Validates the length of a string.
 
@@ -92,35 +116,59 @@ class Length(object):
         Error message to raise in case of a validation error. Can be
         interpolated using `%(min)d` and `%(max)d` if desired. Useful defaults
         are provided depending on the existence of min and max.
+
+    When supported, sets the `minlength` and `maxlength` attributes on widgets.
     """
+
     def __init__(self, min=-1, max=-1, message=None):
-        assert min != -1 or max != -1, 'At least one of `min` or `max` must be specified.'
-        assert max == -1 or min <= max, '`min` cannot be more than `max`.'
+        assert (
+            min != -1 or max != -1
+        ), "At least one of `min` or `max` must be specified."
+        assert max == -1 or min <= max, "`min` cannot be more than `max`."
         self.min = min
         self.max = max
         self.message = message
+        self.field_flags = {}
+        if self.min != -1:
+            self.field_flags["minlength"] = self.min
+        if self.max != -1:
+            self.field_flags["maxlength"] = self.max
 
     def __call__(self, form, field):
-        l = field.data and len(field.data) or 0
-        if l < self.min or self.max != -1 and l > self.max:
+        length = field.data and len(field.data) or 0
+        if length >= self.min and (self.max == -1 or length <= self.max):
+            return
+
+        if self.message is not None:
             message = self.message
-            if message is None:
-                if self.max == -1:
-                    message = field.ngettext('Field must be at least %(min)d character long.',
-                                             'Field must be at least %(min)d characters long.', self.min)
-                elif self.min == -1:
-                    message = field.ngettext('Field cannot be longer than %(max)d character.',
-                                             'Field cannot be longer than %(max)d characters.', self.max)
-                elif self.min == self.max:
-                    message = field.ngettext('Field must be exactly %(max)d character long.',
-                                             'Field must be exactly %(max)d characters long.', self.max)
-                else:
-                    message = field.gettext('Field must be between %(min)d and %(max)d characters long.')
 
-            raise ValidationError(message % dict(min=self.min, max=self.max, length=l))
+        elif self.max == -1:
+            message = field.ngettext(
+                "Field must be at least %(min)d character long.",
+                "Field must be at least %(min)d characters long.",
+                self.min,
+            )
+        elif self.min == -1:
+            message = field.ngettext(
+                "Field cannot be longer than %(max)d character.",
+                "Field cannot be longer than %(max)d characters.",
+                self.max,
+            )
+        elif self.min == self.max:
+            message = field.ngettext(
+                "Field must be exactly %(max)d character long.",
+                "Field must be exactly %(max)d characters long.",
+                self.max,
+            )
+        else:
+            message = field.gettext(
+                "Field must be between %(min)d and %(max)d characters long."
+            )
+
+        raise ValidationError(message % dict(min=self.min, max=self.max, length=length))
 
 
-class NumberRange(object):
+class NumberRange:
     """
     Validates that a number is of a minimum and/or maximum value, inclusive.
     This will work with any comparable number type, such as floats and
@@ -136,31 +184,48 @@ class NumberRange(object):
         Error message to raise in case of a validation error. Can be
         interpolated using `%(min)s` and `%(max)s` if desired. Useful defaults
         are provided depending on the existence of min and max.
+
+    When supported, sets the `min` and `max` attributes on widgets.
     """
+
     def __init__(self, min=None, max=None, message=None):
         self.min = min
         self.max = max
         self.message = message
+        self.field_flags = {}
+        if self.min is not None:
+            self.field_flags["min"] = self.min
+        if self.max is not None:
+            self.field_flags["max"] = self.max
 
     def __call__(self, form, field):
         data = field.data
-        if data is None or math.isnan(data) or (self.min is not None and data < self.min) or \
-                (self.max is not None and data > self.max):
+        if (
+            data is not None
+            and not math.isnan(data)
+            and (self.min is None or data >= self.min)
+            and (self.max is None or data <= self.max)
+        ):
+            return
+
+        if self.message is not None:
             message = self.message
-            if message is None:
-                # we use %(min)s interpolation to support floats, None, and
-                # Decimals without throwing a formatting exception.
-                if self.max is None:
-                    message = field.gettext('Number must be at least %(min)s.')
-                elif self.min is None:
-                    message = field.gettext('Number must be at most %(max)s.')
-                else:
-                    message = field.gettext('Number must be between %(min)s and %(max)s.')
 
-            raise ValidationError(message % dict(min=self.min, max=self.max))
+        # we use %(min)s interpolation to support floats, None, and
+        # Decimals without throwing a formatting exception.
+        elif self.max is None:
+            message = field.gettext("Number must be at least %(min)s.")
+
+        elif self.min is None:
+            message = field.gettext("Number must be at most %(max)s.")
+
+        else:
+            message = field.gettext("Number must be between %(min)s and %(max)s.")
+
+        raise ValidationError(message % dict(min=self.min, max=self.max))
 
 
-class Optional(object):
+class Optional:
     """
     Allows empty input and stops the validation chain from continuing.
 
@@ -170,8 +235,9 @@ class Optional(object):
     :param strip_whitespace:
         If True (the default) also stop the validation chain on input which
         consists of only whitespace.
+
+    Sets the `optional` attribute on widgets.
     """
-    field_flags = ('optional', )
 
     def __init__(self, strip_whitespace=True):
         if strip_whitespace:
@@ -179,13 +245,19 @@ class Optional(object):
         else:
             self.string_check = lambda s: s
 
+        self.field_flags = {"optional": True}
+
     def __call__(self, form, field):
-        if not field.raw_data or isinstance(field.raw_data[0], string_types) and not self.string_check(field.raw_data[0]):
+        if (
+            not field.raw_data
+            or isinstance(field.raw_data[0], str)
+            and not self.string_check(field.raw_data[0])
+        ):
             field.errors[:] = []
             raise StopValidation()
 
 
-class DataRequired(object):
+class DataRequired:
     """
     Checks the field's data is 'truthy' otherwise stops the validation chain.
 
@@ -206,64 +278,56 @@ class DataRequired(object):
 
     :param message:
         Error message to raise in case of a validation error.
+
+    Sets the `required` attribute on widgets.
     """
-    field_flags = ('required', )
 
     def __init__(self, message=None):
         self.message = message
+        self.field_flags = {"required": True}
 
     def __call__(self, form, field):
-        if not field.data or isinstance(field.data, string_types) and not field.data.strip():
-            if self.message is None:
-                message = field.gettext('This field is required.')
-            else:
-                message = self.message
+        if field.data and (not isinstance(field.data, str) or field.data.strip()):
+            return
 
-            field.errors[:] = []
-            raise StopValidation(message)
+        if self.message is None:
+            message = field.gettext("This field is required.")
+        else:
+            message = self.message
+
+        field.errors[:] = []
+        raise StopValidation(message)
 
 
-class Required(DataRequired):
-    """
-    Legacy alias for DataRequired.
-
-    This is needed over simple aliasing for those who require that the
-    class-name of required be 'Required.'
-
-    """
-    def __init__(self, *args, **kwargs):
-        super(Required, self).__init__(*args, **kwargs)
-        warnings.warn(
-            'Required is going away in WTForms 3.0, use DataRequired',
-            DeprecationWarning, stacklevel=2
-        )
-
-
-class InputRequired(object):
+class InputRequired:
     """
     Validates that input was provided for this field.
 
     Note there is a distinction between this and DataRequired in that
     InputRequired looks that form-input data was provided, and DataRequired
     looks at the post-coercion data.
+
+    Sets the `required` attribute on widgets.
     """
-    field_flags = ('required', )
 
     def __init__(self, message=None):
         self.message = message
+        self.field_flags = {"required": True}
 
     def __call__(self, form, field):
-        if not field.raw_data or not field.raw_data[0]:
-            if self.message is None:
-                message = field.gettext('This field is required.')
-            else:
-                message = self.message
+        if field.raw_data and field.raw_data[0]:
+            return
 
-            field.errors[:] = []
-            raise StopValidation(message)
+        if self.message is None:
+            message = field.gettext("This field is required.")
+        else:
+            message = self.message
+
+        field.errors[:] = []
+        raise StopValidation(message)
 
 
-class Regexp(object):
+class Regexp:
     """
     Validates the field against a user provided regexp.
 
@@ -276,33 +340,35 @@ class Regexp(object):
     :param message:
         Error message to raise in case of a validation error.
     """
+
     def __init__(self, regex, flags=0, message=None):
-        if isinstance(regex, string_types):
+        if isinstance(regex, str):
             regex = re.compile(regex, flags)
         self.regex = regex
         self.message = message
 
     def __call__(self, form, field, message=None):
-        match = self.regex.match(field.data or '')
-        if not match:
-            if message is None:
-                if self.message is None:
-                    message = field.gettext('Invalid input.')
-                else:
-                    message = self.message
+        match = self.regex.match(field.data or "")
+        if match:
+            return match
 
-            raise ValidationError(message)
-        return match
+        if message is None:
+            if self.message is None:
+                message = field.gettext("Invalid input.")
+            else:
+                message = self.message
+
+        raise ValidationError(message)
 
 
-class Email(object):
+class Email:
     """
     Validates an email address. Requires email_validator package to be
     installed. For ex: pip install wtforms[email].
 
     :param message:
         Error message to raise in case of a validation error.
-    :param granular_messsage:
+    :param granular_message:
         Use validation failed message from email_validator library
         (Default False).
     :param check_deliverability:
@@ -315,11 +381,6 @@ class Email(object):
         Postfix aliases (Default False).
     """
 
-    user_regex = re.compile(
-        r"(^[-!#$%&'*+/=?^_`{}|~0-9A-Z]+(\.[-!#$%&'*+/=?^_`{}|~0-9A-Z]+)*\Z"  # dot-atom
-        r'|^"([\001-\010\013\014\016-\037!#-\[\]-\177]|\\[\001-\011\013\014\016-\177])*"\Z)',  # quoted-string
-        re.IGNORECASE)
-
     def __init__(
         self,
         message=None,
@@ -328,7 +389,7 @@ class Email(object):
         allow_smtputf8=True,
         allow_empty_local=False,
     ):
-        if email_validator is None:
+        if email_validator is None:  # pragma: no cover
             raise Exception("Install 'email_validator' for email validation support.")
         self.message = message
         self.granular_message = granular_message
@@ -353,12 +414,12 @@ class Email(object):
                     message = field.gettext(e)
                 else:
                     message = field.gettext("Invalid email address.")
-            raise ValidationError(message)
+            raise ValidationError(message) from e
 
 
-class IPAddress(object):
+class IPAddress:
     """
-    Validates an IP address. Requires ipaddress package to be instaled for Python 2 support.
+    Validates an IP address.
 
     :param ipv4:
         If True, accept IPv4 addresses as valid (default True)
@@ -367,11 +428,12 @@ class IPAddress(object):
     :param message:
         Error message to raise in case of a validation error.
     """
+
     def __init__(self, ipv4=True, ipv6=False, message=None):
-        if ipaddress is None:
-            raise Exception("Install 'ipaddress' for Python 2 support.")
         if not ipv4 and not ipv6:
-            raise ValueError('IP Address Validator must have at least one of ipv4 or ipv6 enabled.')
+            raise ValueError(
+                "IP Address Validator must have at least one of ipv4 or ipv6 enabled."
+            )
         self.ipv4 = ipv4
         self.ipv6 = ipv6
         self.message = message
@@ -380,13 +442,17 @@ class IPAddress(object):
         value = field.data
         valid = False
         if value:
-            valid = (self.ipv4 and self.check_ipv4(value)) or (self.ipv6 and self.check_ipv6(value))
+            valid = (self.ipv4 and self.check_ipv4(value)) or (
+                self.ipv6 and self.check_ipv6(value)
+            )
 
-        if not valid:
-            message = self.message
-            if message is None:
-                message = field.gettext('Invalid IP address.')
-            raise ValidationError(message)
+        if valid:
+            return
+
+        message = self.message
+        if message is None:
+            message = field.gettext("Invalid IP address.")
+        raise ValidationError(message)
 
     @classmethod
     def check_ipv4(cls, value):
@@ -420,16 +486,17 @@ class MacAddress(Regexp):
     :param message:
         Error message to raise in case of a validation error.
     """
+
     def __init__(self, message=None):
-        pattern = r'^(?:[0-9a-fA-F]{2}:){5}[0-9a-fA-F]{2}$'
-        super(MacAddress, self).__init__(pattern, message=message)
+        pattern = r"^(?:[0-9a-fA-F]{2}:){5}[0-9a-fA-F]{2}$"
+        super().__init__(pattern, message=message)
 
     def __call__(self, form, field):
         message = self.message
         if message is None:
-            message = field.gettext('Invalid Mac address.')
+            message = field.gettext("Invalid Mac address.")
 
-        super(MacAddress, self).__call__(form, field, message)
+        super().__call__(form, field, message)
 
 
 class URL(Regexp):
@@ -445,6 +512,7 @@ class URL(Regexp):
     :param message:
         Error message to raise in case of a validation error.
     """
+
     def __init__(self, require_tld=True, message=None):
         regex = (
             r"^[a-z]+://"
@@ -453,43 +521,43 @@ class URL(Regexp):
             r"(?P<path>\/.*?)?"
             r"(?P<query>\?.*)?$"
         )
-        super(URL, self).__init__(regex, re.IGNORECASE, message)
+        super().__init__(regex, re.IGNORECASE, message)
         self.validate_hostname = HostnameValidation(
-            require_tld=require_tld,
-            allow_ip=True,
+            require_tld=require_tld, allow_ip=True
         )
 
     def __call__(self, form, field):
         message = self.message
         if message is None:
-            message = field.gettext('Invalid URL.')
+            message = field.gettext("Invalid URL.")
 
-        match = super(URL, self).__call__(form, field, message)
-        if not self.validate_hostname(match.group('host')):
+        match = super().__call__(form, field, message)
+        if not self.validate_hostname(match.group("host")):
             raise ValidationError(message)
 
 
-class UUID(object):
+class UUID:
     """
     Validates a UUID.
 
     :param message:
         Error message to raise in case of a validation error.
     """
+
     def __init__(self, message=None):
         self.message = message
 
     def __call__(self, form, field):
         message = self.message
         if message is None:
-            message = field.gettext('Invalid UUID.')
+            message = field.gettext("Invalid UUID.")
         try:
             uuid.UUID(field.data)
-        except ValueError:
-            raise ValidationError(message)
+        except ValueError as exc:
+            raise ValidationError(message) from exc
 
 
-class AnyOf(object):
+class AnyOf:
     """
     Compares the incoming data to a sequence of valid inputs.
 
@@ -501,6 +569,7 @@ class AnyOf(object):
     :param values_formatter:
         Function used to format the list of values in the error message.
     """
+
     def __init__(self, values, message=None, values_formatter=None):
         self.values = values
         self.message = message
@@ -509,19 +578,21 @@ class AnyOf(object):
         self.values_formatter = values_formatter
 
     def __call__(self, form, field):
-        if field.data not in self.values:
-            message = self.message
-            if message is None:
-                message = field.gettext('Invalid value, must be one of: %(values)s.')
+        if field.data in self.values:
+            return
 
-            raise ValidationError(message % dict(values=self.values_formatter(self.values)))
+        message = self.message
+        if message is None:
+            message = field.gettext("Invalid value, must be one of: %(values)s.")
+
+        raise ValidationError(message % dict(values=self.values_formatter(self.values)))
 
     @staticmethod
     def default_values_formatter(values):
-        return ', '.join(text_type(x) for x in values)
+        return ", ".join(str(x) for x in values)
 
 
-class NoneOf(object):
+class NoneOf:
     """
     Compares the incoming data to a sequence of invalid inputs.
 
@@ -533,6 +604,7 @@ class NoneOf(object):
     :param values_formatter:
         Function used to format the list of values in the error message.
     """
+
     def __init__(self, values, message=None, values_formatter=None):
         self.values = values
         self.message = message
@@ -541,60 +613,63 @@ class NoneOf(object):
         self.values_formatter = values_formatter
 
     def __call__(self, form, field):
-        if field.data in self.values:
-            message = self.message
-            if message is None:
-                message = field.gettext('Invalid value, can\'t be any of: %(values)s.')
+        if field.data not in self.values:
+            return
 
-            raise ValidationError(message % dict(values=self.values_formatter(self.values)))
+        message = self.message
+        if message is None:
+            message = field.gettext("Invalid value, can't be any of: %(values)s.")
+
+        raise ValidationError(message % dict(values=self.values_formatter(self.values)))
 
     @staticmethod
     def default_values_formatter(v):
-        return ', '.join(text_type(x) for x in v)
+        return ", ".join(str(x) for x in v)
 
 
-class HostnameValidation(object):
+class HostnameValidation:
     """
     Helper class for checking hostnames for validation.
 
     This is not a validator in and of itself, and as such is not exported.
     """
-    hostname_part = re.compile(r'^(xn-|[a-z0-9_]+)(-[a-z0-9_]+)*$', re.IGNORECASE)
-    tld_part = re.compile(r'^([a-z]{2,20}|xn--([a-z0-9]+-)*[a-z0-9]+)$', re.IGNORECASE)
+
+    hostname_part = re.compile(r"^(xn-|[a-z0-9_]+)(-[a-z0-9_-]+)*$", re.IGNORECASE)
+    tld_part = re.compile(r"^([a-z]{2,20}|xn--([a-z0-9]+-)*[a-z0-9]+)$", re.IGNORECASE)
 
     def __init__(self, require_tld=True, allow_ip=False):
         self.require_tld = require_tld
         self.allow_ip = allow_ip
 
     def __call__(self, hostname):
-        if self.allow_ip:
-            if IPAddress.check_ipv4(hostname) or IPAddress.check_ipv6(hostname):
-                return True
+        if self.allow_ip and (
+            IPAddress.check_ipv4(hostname) or IPAddress.check_ipv6(hostname)
+        ):
+            return True
 
         # Encode out IDNA hostnames. This makes further validation easier.
         try:
-            hostname = hostname.encode('idna')
+            hostname = hostname.encode("idna")
         except UnicodeError:
             pass
 
         # Turn back into a string in Python 3x
-        if not isinstance(hostname, string_types):
-            hostname = hostname.decode('ascii')
+        if not isinstance(hostname, str):
+            hostname = hostname.decode("ascii")
 
         if len(hostname) > 253:
             return False
 
         # Check that all labels in the hostname are valid
-        parts = hostname.split('.')
+        parts = hostname.split(".")
         for part in parts:
             if not part or len(part) > 63:
                 return False
             if not self.hostname_part.match(part):
                 return False
 
-        if self.require_tld:
-            if len(parts) < 2 or not self.tld_part.match(parts[-1]):
-                return False
+        if self.require_tld and (len(parts) < 2 or not self.tld_part.match(parts[-1])):
+            return False
 
         return True
 
@@ -606,7 +681,6 @@ mac_address = MacAddress
 length = Length
 number_range = NumberRange
 optional = Optional
-required = Required
 input_required = InputRequired
 data_required = DataRequired
 regexp = Regexp
